#include <pioneer_module/pioneer_module.h>

CPioneerModule::CPioneerModule(const std::string &name,const std::string &name_space) : CModule(name,name_space),
  //SERVICES
  enable_motors_service("enable_motors",this->module_nh.getNamespace()),
  disable_motors_service("disable_motors",this->module_nh.getNamespace()),
  pioneer_reconf("pioneer_reconf",this->module_nh.getNamespace())
{

  this->start_operation();
  /****************Topics*****************/
  this->battery_voltage_subscriber = this->module_nh.subscribe("battery_voltage",1,&CPioneerModule::battery_voltage_callback,this);
  this->battery_watchdog.reset(ros::Duration(this->config.battery_watchdog));
  this->battery_voltage=0.0;
  this->motors_state_subscriber = this->module_nh.subscribe("motors_state",1,&CPioneerModule::motors_state_callback,this);
  this->state_watchdog.reset(ros::Duration(this->config.state_watchdog));
  this->motors_enabled=false;

  /****************Actions******************/
  this->status=PIONEER_MODULE_SUCCESS;
}

/* State machine */
void CPioneerModule::state_machine(void)
{
}

void CPioneerModule::battery_voltage_callback(const std_msgs::Float64::ConstPtr& msg)
{
  this->lock();
  this->battery_watchdog.reset(ros::Duration(this->config.battery_watchdog));
  this->battery_voltage=msg->data;
  this->unlock();
}

void CPioneerModule::motors_state_callback(const std_msgs::Bool::ConstPtr& msg)
{
  this->lock();
  this->state_watchdog.reset(ros::Duration(this->config.state_watchdog));
  this->motors_enabled=msg->data;
  this->unlock();
}

bool CPioneerModule::enable_motors(void)
{
  std_srvs::Empty empty_req;

  this->lock();
  switch(this->enable_motors_service.call(empty_req))
  {
    case ACT_SRV_SUCCESS: ROS_INFO("CPioneerModule: Motors enabled successfully");
                          this->status=PIONEER_MODULE_SUCCESS;
                          this->unlock();
                          return true;
    case ACT_SRV_PENDING: ROS_WARN("CPioneerModule: Motors not yet enabled");
                          this->status=PIONEER_MODULE_ENABLE_PENDING;
                          this->unlock();
                          return false;
                          break;
    case ACT_SRV_FAIL: ROS_ERROR("CPioneerModule: Impossible to enable motors");
                       this->status=PIONEER_MODULE_ENABLE_FAILED;
                       this->unlock();
                       return false;
                       break;
  }

  this->unlock();
  return false;
}

bool CPioneerModule::disable_motors(void)
{
  std_srvs::Empty empty_req;

  this->lock();
  switch(this->disable_motors_service.call(empty_req))
  {
    case ACT_SRV_SUCCESS: ROS_INFO("CPioneerModule: Motors disabled successfully");
                          this->status=PIONEER_MODULE_SUCCESS;
                          this->unlock();
                          return true;
    case ACT_SRV_PENDING: ROS_WARN("CPioneerModule: Motors not yet disabled");
                          this->status=PIONEER_MODULE_DISABLE_PENDING;
                          this->unlock();
                          return false;
                          break;
    case ACT_SRV_FAIL: ROS_ERROR("CPioneerModule: Impossible to disable motors");
                       this->status=PIONEER_MODULE_DISABLE_FAILED;
                       this->unlock();
                       return false;
                       break;
  }

  this->unlock();
  return false;
}

bool CPioneerModule::are_motors_enabled(bool &enabled)
{
  this->lock();
  if(this->state_watchdog.is_active())
  {
    this->status=PIONEER_MODULE_STATE_WATCHDOG;
    this->unlock();
    return false;
  }
  else
  {
    this->status=PIONEER_MODULE_SUCCESS;
    enabled=this->motors_enabled;
    this->unlock();
    return true; 
  }
}

bool CPioneerModule::get_battery_voltage(double &voltage)
{
  this->lock();
  if(this->battery_watchdog.is_active())
  {
    this->status=PIONEER_MODULE_BATTERY_WATCHDOG;
    this->unlock();
    return false;
  }
  else
  {
    this->status=PIONEER_MODULE_SUCCESS;
    voltage=this->battery_voltage;
    this->unlock();
    return true; 
  }
}

bool CPioneerModule::set_trans_accel(float accel_m_s_s)
{
  this->lock();
  if(!this->pioneer_reconf.set_parameter("trans_accel",accel_m_s_s))
  {
    if(this->pioneer_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status=PIONEER_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist, ignore it and read the current position from odom
    {
      this->status=PIONEER_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  else
  {
    this->unlock();
    return true;
  }
}

bool CPioneerModule::set_trans_decel(float decel_m_s_s)
{
  this->lock();
  if(!this->pioneer_reconf.set_parameter("trans_decel",decel_m_s_s))
  {
    if(this->pioneer_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status=PIONEER_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist, ignore it and read the current position from odom
    {
      this->status=PIONEER_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  else
  {
    this->unlock();
    return true;
  }
}

bool CPioneerModule::set_rot_accel(float accel_rad_s)
{
  this->lock();
  if(!this->pioneer_reconf.set_parameter("rot_accel",accel_rad_s))
  {
    if(this->pioneer_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status=PIONEER_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist, ignore it and read the current position from odom
    {
      this->status=PIONEER_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  else
  {
    this->unlock();
    return true;
  }
}

bool CPioneerModule::set_rot_devel(float decel_rad_s)
{
  this->lock();
  if(!this->pioneer_reconf.set_parameter("rot_decel",decel_rad_s))
  {
    if(this->pioneer_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status=PIONEER_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist, ignore it and read the current position from odom
    {
      this->status=PIONEER_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  else
  {
    this->unlock();
    return true;
  }
}

bool CPioneerModule::set_cmd_vel_timeout(float time_s)
{
  this->lock();
  if(!this->pioneer_reconf.set_parameter("cmd_vel_timeout",time_s))
  {
    if(this->pioneer_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status=PIONEER_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist, ignore it and read the current position from odom
    {
      this->status=PIONEER_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  else
  {
    this->unlock();
    return true;
  }
}


void CPioneerModule::reconfigure_callback(pioneer_module::PioneerModuleConfig &config, uint32_t level)
{
  ROS_INFO("CPioneerModule : reconfigure callback");
  this->lock();
  this->config=config;
  /* set the module rate */
  this->set_rate(config.rate_hz);
  /* set enable motors service parameters */
  this->enable_motors_service.set_max_num_retries(config.enable_max_retries);
  if(this->config.enable_enabled)
    this->enable_motors_service.enable();
  else
    this->enable_motors_service.disable();
  /* set disable motors service parameters */
  this->disable_motors_service.set_max_num_retries(config.disable_max_retries);
  if(this->config.disable_enabled)
    this->disable_motors_service.enable();
  else
    this->disable_motors_service.disable();

  this->unlock();
}

pioneer_module_status_t CPioneerModule::get_status(void)
{
  return this->status;
}

CPioneerModule::~CPioneerModule()
{
}
